/* 
 *  mod_authn_mysql.c -- MySQL authentication module for Apache 2.2/2.4
 *
 *  Copyright (C) Svein Olav Bjerkeset, 2012
 *  License: GNU GENERAL PUBLIC LICENSE Version 3
 */

#include <sys/types.h>
#include <unistd.h>
#ifdef SOLARIS2
#include <crypt.h>
#endif
#include <string.h>

#include <apr_strings.h>
#include <ap_provider.h>
#include <ap_config.h>

#include <httpd.h>
#include <http_config.h>
#include <http_core.h>
#include <http_log.h>
#include <http_protocol.h>
#include <http_request.h>

#include <mod_auth.h>

#include <mysql.h>

#include "config.h"
#include "myauth_config.h"

module AP_MODULE_DECLARE_DATA authn_mysql_module;


/* get password for user from database */
static char *get_pw (request_rec *r, myauth_config_rec *cfg)
{
	MYSQL		*dbh = NULL;
	MYSQL_RES	*res = NULL;
	MYSQL_ROW	row;
	char		*sql, *pw = NULL;

	if (!(dbh = mysql_init(NULL))) {
		ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, r->server, 
			"authn_mysql: cannot initialize MySQL");
		return NULL;
	}
	
	if (!mysql_real_connect(dbh, cfg->my_host, cfg->my_login, cfg->my_passw,
		cfg->my_db, cfg->my_port, NULL, CLIENT_COMPRESS)) {
		ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, r->server, 
			"authn_mysql: connection failed: %s", mysql_error(dbh));
		mysql_close(dbh);
		return NULL;
	}
	
	sql = apr_psprintf(r->pool, 
		"SELECT %s FROM %s WHERE %s = '%s'", 
		cfg->my_pwfld, cfg->my_usertbl, cfg->my_userfld, r->user);
	if (mysql_query(dbh, sql)) {
		ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, r->server, 
			"authn_mysql: query failed: %s", mysql_error(dbh));
		mysql_close(dbh);
		return NULL;
	}
	if (!(res = mysql_store_result(dbh))) {
		ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, r->server, 
			"authn_mysql: could not store query result: %s", mysql_error(dbh));
		mysql_close(dbh);
		return NULL;
	}
	
	switch (mysql_num_rows(res)) {
	case 1:
		if (!(row = mysql_fetch_row(res))) {
			ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, r->server, 
				"authn_mysql: could not get row: %s", mysql_error(dbh));
			break;
		}
		pw = apr_pstrdup (r->pool, row[0]);
		break;
	case 0:
		break;
	default:
		ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, r->server, 
			"authn_mysql: user '%s' not unique", r->user);
		break;
	}
	
	mysql_free_result(res);
	mysql_close(dbh);
	return pw;
}


#ifdef USE_OPENSSL
/**
  Database and OS independent encryption: Create an MD5 digest from
  msg, Base64 encode the obtained digest and store the result in buf.

  @param msg	string to encode
  @param buf	where to store the null terminated result of encoding.
  				If buflen is actually smaller than the result, the result
				is truncated, but still null-terminated.
  @param buflen	max. length of buf (including null-terminator)

  @return  the length of the encoded string.
*/
static int md5base64(const char *msg, char *buf, int buflen)
{
    EVP_MD_CTX mdctx;
    const EVP_MD *md;
    unsigned char md_value[EVP_MAX_MD_SIZE];
    int md_len, i;
    BIO *mem, *bio, *b64;

    memset(buf, 0, buflen);

    /* digest md5 */
    md = EVP_md5();
    if(!md) {
        return 0;
    }
    EVP_MD_CTX_init(&mdctx);
    EVP_DigestInit_ex(&mdctx, md, NULL);
    EVP_DigestUpdate(&mdctx, msg, strlen(msg));
    EVP_DigestFinal_ex(&mdctx, md_value, &md_len);
    EVP_MD_CTX_cleanup(&mdctx);

    /* base64 encoding */
    b64 = BIO_new(BIO_f_base64());
    mem = BIO_new(BIO_s_mem());
    bio = BIO_push(b64, mem);

    BIO_write(bio, md_value, md_len);
    BIO_flush(bio);

    i = BIO_read(mem, buf, buflen);
    if (i > 0) {
        i--; /* kill the newline char */
    } else {
       i = 0;
    }
    buf[i] = '\0';
    BIO_free_all(bio);
    return i;
}
#endif


static int valid_pw (request_rec *r, const char *sent, const char *real, const int etyp)
{
	char my_pw[64];

	ap_log_rerror (APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, 0, r, 
		"encryption mask = 0x%x", etyp);
#ifdef USE_OPENSSL
	if (etyp & MYAUTH_ENC_MD5BASE64) {
		int i = md5base64(sent, my_pw, sizeof(my_pw));
        if (i > 0 && strcmp(real, my_pw) == 0)
            return 1;
    }
#endif
	if (etyp & MYAUTH_ENC_MYSQL) {
#ifdef HAVE_MAKE_SCRAMBLED_PASSWORD
		make_scrambled_password (my_pw, sent);
		if (!strcmp (real, my_pw))
			return 1;
#endif
#ifdef HAVE_MAKE_SCRAMBLED_PASSWORD_323
		make_scrambled_password_323 (my_pw, sent);
		if (!strcmp (real, my_pw))
			return 1;
#endif
	}
	if ((etyp & MYAUTH_ENC_CRYPT) && !strcmp (real,crypt(sent,real)))
		return 1;
	if ((etyp & MYAUTH_ENC_PLAIN) && !strcmp (real,sent))
		return 1;
	return 0;
}


/* Authenticate user */
static authn_status authn_mysql_verify_password (request_rec *r, const char *user, const char *password)
{
	myauth_config_rec *conf =
		ap_get_module_config(r->per_dir_config, &authn_mysql_module);
	char *real_pw;
	int res;
	
	ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, 0, r,
		"authn_mysql: trying to authenticate user '%s'", r->user);

	if (!conf->my_db || !conf->my_login)
		return AUTH_GENERAL_ERROR;

	if (!(real_pw = get_pw(r, conf))) {
		if (!(conf->my_authoritative))
			return AUTH_USER_NOT_FOUND;
		ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, r,
			"authn_mysql: user '%s' not found: %s", r->user, r->uri);
		ap_note_basic_auth_failure(r);
		return AUTH_USER_NOT_FOUND;
	}

	if (!valid_pw (r, password, real_pw, conf->my_enctype)) {
		ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, r,
			"authn_mysql: user '%s' authentication failure for '%s': "
			"incorrect password",
			r->user, r->uri);
		ap_note_basic_auth_failure(r);
		return AUTH_DENIED;
	}
	return AUTH_GRANTED;
}


/* Module initialization handler */
static int authn_mysql_init_handler(apr_pool_t *pconf, apr_pool_t *plog, apr_pool_t *ptemp, server_rec *s)
{
	ap_add_version_component(pconf, "mod_authn_mysql/2.0");
	return OK;
}


static const authn_provider authn_mysql_provider =
{
	&authn_mysql_verify_password,
	NULL
};


static void authn_mysql_register_hooks (apr_pool_t *p)
{
	ap_hook_post_config (authn_mysql_init_handler, NULL, NULL, APR_HOOK_MIDDLE);
	ap_register_auth_provider (p, AUTHN_PROVIDER_GROUP, "mysql",
		AUTHN_PROVIDER_VERSION,
		&authn_mysql_provider, AP_AUTH_INTERNAL_PER_CONF);
}


static void *authn_mysql_init_config (apr_pool_t *p, char *d)
{
	myauth_config_rec *cfg = (myauth_config_rec*) 
		apr_pcalloc(p, sizeof(myauth_config_rec));

	/* default values */
	cfg->my_host          = apr_pstrdup (p, "localhost");
	cfg->my_port          = 0;
	cfg->my_db            = NULL;
	cfg->my_login         = NULL;
	cfg->my_passw         = NULL;
	cfg->my_authoritative = 1;
	cfg->my_usertbl       = apr_pstrdup (p, "users");
	cfg->my_userfld       = apr_pstrdup (p, "name");
	cfg->my_pwfld         = apr_pstrdup (p, "password");
	cfg->my_enctype       = MYAUTH_ENC_MYSQL;

	return cfg;
}


static const command_rec authn_mysql_commands[] = {
	AP_INIT_TAKE1( "MyAuthHost", myauth_set_slot,
		(void *) APR_OFFSETOF(myauth_config_rec, my_host),
		OR_AUTHCFG, "host to connect to." ),
	AP_INIT_TAKE1("MyAuthPort", myauth_set_port,
		NULL,
		OR_AUTHCFG, "port to connect on." ),
	AP_INIT_TAKE1("MyAuthDB", myauth_set_slot, 
		(void *) APR_OFFSETOF (myauth_config_rec, my_db),
		OR_AUTHCFG, "database to use for authentication." ),
	AP_INIT_TAKE1("MyAuthDBUser", myauth_set_slot, 
		(void *) APR_OFFSETOF (myauth_config_rec, my_login),
		OR_AUTHCFG, "username to use for connection." ),
	AP_INIT_TAKE1("MyAuthDBPassw", myauth_set_slot,
		(void *) APR_OFFSETOF (myauth_config_rec, my_passw),
		OR_AUTHCFG, "password to use for connection." ),
	AP_INIT_FLAG("MyAuthAuthoritative",  ap_set_flag_slot,
		(void *) APR_OFFSETOF (myauth_config_rec, my_authoritative),
		OR_AUTHCFG, "set authoritative on or off." ), 
	AP_INIT_TAKE1("MyAuthUserTable", myauth_set_slot,	
		(void *) APR_OFFSETOF(myauth_config_rec, my_usertbl),
		OR_AUTHCFG, "table to use for authentication." ),
	AP_INIT_TAKE1("MyAuthUserField", myauth_set_slot,
		(void *) APR_OFFSETOF (myauth_config_rec, my_userfld),
		OR_AUTHCFG, "user field in table." ),
	AP_INIT_TAKE1("MyAuthPasswordField", myauth_set_slot,
		(void *) APR_OFFSETOF (myauth_config_rec, my_pwfld),
		OR_AUTHCFG, "password field in table." ),
	AP_INIT_TAKE12 ("MyAuthEncryption", myauth_set_enctype,	
		NULL,
		OR_AUTHCFG, "set encryption method(s)." ),
	{ NULL }
};


/* Dispatch list for API hooks */
module AP_MODULE_DECLARE_DATA authn_mysql_module = {
	STANDARD20_MODULE_STUFF, 
	authn_mysql_init_config,    /* create per-dir    config structures */
	NULL,                       /* merge  per-dir    config structures */
	NULL,                       /* create per-server config structures */
	NULL,                       /* merge  per-server config structures */
	authn_mysql_commands,       /* table of config file commands       */
	authn_mysql_register_hooks  /* register hooks                      */
};
