/* 
**  mod_authz_mysql.c -- MySQL authorization module for Apache 2.2/2.4
**
**  Copyright (C) Svein Olav Bjerkeset, January 2012
**  License: GNU GENERAL PUBLIC LICENSE Version 3
*/

#include <sys/types.h>
#include <unistd.h>
#include <string.h>

#include <ap_provider.h>
#include <ap_config.h>

#include <httpd.h>
#include <http_config.h>
#include <http_core.h>
#include <http_log.h>
#include <http_protocol.h>
#include <http_request.h>

#include <mod_auth.h>

#include <mysql.h>

#include "config.h"
#include "myauth_config.h"

module AP_MODULE_DECLARE_DATA authz_mysql_module;


static apr_status_t get_groups (request_rec *r, myauth_config_rec *cfg,
	apr_array_header_t **out)
{
	MYSQL		*dbh = NULL;
	MYSQL_RES	*res = NULL;
	MYSQL_ROW	row;
	char		*sql, *pw = NULL;

	if (!(dbh = mysql_init(NULL))) {
		ap_log_error (APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, r->server, 
			"authz_mysql: cannot initialize MySQL");
		return APR_EGENERAL;
	}
	
	if (!mysql_real_connect(dbh, cfg->my_host, cfg->my_login, cfg->my_passw,
		cfg->my_db, cfg->my_port, NULL, CLIENT_COMPRESS)) {
		ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, r->server, 
			"authz_mysql: connection failed: %s", mysql_error(dbh));
		mysql_close(dbh);
		return APR_EGENERAL;
	}
	
	sql = apr_psprintf(r->pool, 
		"SELECT %s FROM %s WHERE %s = '%s'", 
		cfg->my_grpfld, cfg->my_grptbl, cfg->my_userfld, r->user);
	if (mysql_query(dbh, sql)) {
		ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, r->server, 
			"authz_mysql: group query failed: %s", mysql_error(dbh));
		mysql_close(dbh);
		return APR_EGENERAL;
	}
	if (!(res = mysql_store_result(dbh))) {
		ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, r->server, 
			"authz_mysql: could not store group query result: %s", mysql_error(dbh));
		mysql_close(dbh);
		return APR_EGENERAL;
	}
	
	while (row = mysql_fetch_row(res)) {
		char *group_name = row[0];
		*(const char **)apr_array_push (*out) = apr_pstrdup(r->pool, group_name);
	}
	
	mysql_free_result(res);
	mysql_close(dbh);
	return APR_SUCCESS;
}


/* Check for valid user */
static authz_status authz_mysql_check_authorization (request_rec *r,
	const char* require_args, const void *parsed_require_args)
{
	char *user = r->user;
	const char *t;
	char *w;
	int i;
	apr_array_header_t *groups = apr_array_make (r->pool, 10, sizeof (char *));
	
	if (!user)
		return AUTHZ_DENIED_NO_USER;
		
	if (!groups)
		return AUTHZ_GENERAL_ERROR;
	
	/* get groups for user */
	myauth_config_rec *conf = 
		ap_get_module_config (r->per_dir_config, &authz_mysql_module);
	if (get_groups (r, conf, &groups) != APR_SUCCESS) {
		ap_log_rerror (APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, r, 
			"authz_mysql: could not query groups for user '%s'", user);
		return AUTHZ_GENERAL_ERROR;
	}
	
	/* if user has no group memberships, we are done */
	if (groups->nelts <= 0) {
		ap_log_rerror (APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, r, 
			"authz_mysql: access denied: user '%s' has no group membership", user);
		return AUTHZ_DENIED;
	}
	
	/* traverse all groups */
	t = require_args;
	while ((w = ap_getword_white (r->pool, &t)) && w[0]) {
		for (i=0; i<groups->nelts; i++) {
			const char *member_group = ((const char**)groups->elts)[i];
			if (!(strcmp (w, member_group)))
				return AUTHZ_GRANTED;
		}
	}
	
	/* user is not member of any required group, access is denied */
	ap_log_rerror (APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, r, 
		"authz_mysql: access denied: user '%s' not member of any required group", user);
	return AUTHZ_DENIED;
}


/* module initialization handler */
static int authz_mysql_init_handler (apr_pool_t *pconf, apr_pool_t *plog, apr_pool_t *ptemp, server_rec *s)
{
	ap_add_version_component (pconf, "mod_authz_mysql/2.0");
	return OK;
}


/* provider definition */
static const authz_provider authz_mysql_provider =
{
	&authz_mysql_check_authorization,
	NULL
};


/* module hooks */
static void authz_mysql_register_hooks (apr_pool_t *p)
{
	ap_hook_post_config (authz_mysql_init_handler, NULL, NULL, APR_HOOK_MIDDLE);
	ap_register_auth_provider (p, AUTHZ_PROVIDER_GROUP, "mysql-group",
		AUTHZ_PROVIDER_VERSION, &authz_mysql_provider, AP_AUTH_INTERNAL_PER_CONF);
}


static void *authz_mysql_init_config (apr_pool_t *p, char *d)
{
	myauth_config_rec *cfg = (myauth_config_rec*) 
		apr_pcalloc(p, sizeof(myauth_config_rec));

	/* default values */
	cfg->my_host          = apr_pstrdup (p, "localhost");
	cfg->my_port          = 0;
	cfg->my_db            = NULL;
	cfg->my_login         = NULL;
	cfg->my_passw         = NULL;
	cfg->my_authoritative = 1;
	cfg->my_usertbl       = apr_pstrdup (p, "users");
	cfg->my_userfld       = apr_pstrdup (p, "name");
	cfg->my_grptbl        = apr_pstrdup (p, "groups");
	cfg->my_grpfld        = apr_pstrdup (p, "gname");

	return cfg;
}


static const command_rec authz_mysql_commands[] = {
	AP_INIT_TAKE1( "MyAuthHost", myauth_set_slot,
		(void *) APR_OFFSETOF(myauth_config_rec, my_host),
		OR_AUTHCFG, "host to connect to." ),
	AP_INIT_TAKE1("MyAuthPort", myauth_set_port,
		NULL,
		OR_AUTHCFG, "port to connect on." ),
	AP_INIT_TAKE1("MyAuthDB", myauth_set_slot, 
		(void *) APR_OFFSETOF (myauth_config_rec, my_db),
		OR_AUTHCFG, "database to use for authentication." ),
	AP_INIT_TAKE1("MyAuthDBUser", myauth_set_slot, 
		(void *) APR_OFFSETOF (myauth_config_rec, my_login),
		OR_AUTHCFG, "username to use for connection." ),
	AP_INIT_TAKE1("MyAuthDBPassw", myauth_set_slot,
		(void *) APR_OFFSETOF (myauth_config_rec, my_passw),
		OR_AUTHCFG, "password to use for connection." ),
	AP_INIT_FLAG("MyAuthAuthoritative",  ap_set_flag_slot,
		(void *) APR_OFFSETOF (myauth_config_rec, my_authoritative),
		OR_AUTHCFG, "set authoritative on or off." ), 
	AP_INIT_TAKE1("MyAuthUserTable", myauth_set_slot,	
		(void *) APR_OFFSETOF(myauth_config_rec, my_usertbl),
		OR_AUTHCFG, "table to use for authentication." ),
	AP_INIT_TAKE1("MyAuthUserField", myauth_set_slot,
		(void *) APR_OFFSETOF (myauth_config_rec, my_userfld),
		OR_AUTHCFG, "user field in table." ),
	AP_INIT_TAKE1 ("MyAuthGroupTable", myauth_set_slot,
		(void *) APR_OFFSETOF (myauth_config_rec, my_grptbl),
		OR_AUTHCFG, "table holding group memberships." ),
	AP_INIT_TAKE1 ("MyAuthGroupField", myauth_set_slot,
		(void *) APR_OFFSETOF (myauth_config_rec, my_grpfld),
		OR_AUTHCFG, "name of group name column." ),
	{ NULL }
};


/* dispatch list for API hooks */
module AP_MODULE_DECLARE_DATA authz_mysql_module = {
    STANDARD20_MODULE_STUFF, 
    authz_mysql_init_config,    /* create per-dir    config structures */
    NULL,                       /* merge  per-dir    config structures */
    NULL,                       /* create per-server config structures */
    NULL,                       /* merge  per-server config structures */
    authz_mysql_commands,       /* table of config file commands       */
    authz_mysql_register_hooks  /* register hooks                      */
};
